#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
import re
from collections import Counter

from tqdm import tqdm

from dan.datasets.format.generic import OCRDatasetFormatter


def remove_spaces(text):
    # remove begin/ending spaces
    text = text.strip()
    # replace \t with regular space
    text = re.sub("\t", " ", text)
    # remove consecutive spaces
    text = re.sub(" +", " ", text)
    #    text = text.encode('ascii', 'ignore').decode("utf-8")
    return text


class BessinDatasetFormatter(OCRDatasetFormatter):
    def __init__(
        self, level, set_names=["train", "valid", "test"], dpi=150, sem_token=False
    ):
        super(BessinDatasetFormatter, self).__init__(
            "bessin", level, "_sem" if sem_token else "", set_names
        )

        self.dpi = dpi
        self.counter = Counter()
        self.map_datasets_files.update(
            {
                "bessin": {
                    # (1,050 for train, 100 for validation and 100 for test)
                    "line": {
                        "needed_files": [],
                        "arx_files": [],
                        "format_function": self.format_bessin_zone,
                    }
                }
            }
        )

    def preformat_bessin_zone(self):
        """
        Extract all information from dataset and correct some annotations
        """

        dataset = {"train": list(), "valid": list(), "test": list()}
        img_folder_path = os.path.join("Datasets", "raw", "bessin", "images")
        labels_folder_path = os.path.join("Datasets", "raw", "bessin", "labels")

        train_files = [
            os.path.join(labels_folder_path, "train", name)
            for name in os.listdir(os.path.join(labels_folder_path, "train"))
        ]
        valid_files = [
            os.path.join(labels_folder_path, "valid", name)
            for name in os.listdir(os.path.join(labels_folder_path, "valid"))
        ]
        test_files = [
            os.path.join(labels_folder_path, "test", name)
            for name in os.listdir(os.path.join(labels_folder_path, "test"))
        ]

        for set_name, files in zip(
            self.set_names, [train_files, valid_files, test_files]
        ):
            for i, label_file in enumerate(
                tqdm(files, desc="Pre-formatting " + set_name)
            ):
                with open(label_file, "r") as f:
                    text = remove_spaces(f.read())

                dataset[set_name].append(
                    {
                        "img_path": os.path.join(
                            img_folder_path,
                            set_name,
                            label_file.split("/")[-1].replace("txt", "jpg"),
                        ),
                        "label": text.strip(),
                    }
                )

        return dataset

    def format_bessin_zone(self):
        """
        Format synist page dataset
        """
        dataset = self.preformat_bessin_zone()
        for set_name in self.set_names:
            fold = os.path.join(self.target_fold_path, set_name)
            for sample in tqdm(dataset[set_name], desc="Formatting " + set_name):
                new_name = sample["img_path"].split("/")[-1]
                new_img_path = os.path.join(fold, new_name)
                self.load_resize_save(sample["img_path"], new_img_path)
                zone = {
                    "text": sample["label"],
                }
                self.charset = self.charset.union(set(zone["text"]))
                self.gt[set_name][new_name] = zone
                self.counter.update(zone["text"])


if __name__ == "__main__":
    formatter = BessinDatasetFormatter("line", sem_token=False)
    formatter.format()
    print("Character freq: ")
    for k, v in formatter.counter.items():
        print(k, v)
