# -*- coding: utf-8 -*-
import numpy as np

from dan.datasets.format.simara import SEM_MATCHING_TOKENS as SIMARA_MATCHING_TOKENS


class PostProcessingModule:
    """
    Forward pass post processing
    Add/remove layout tokens only to:
     - respect token hierarchy
     - complete/remove unpaired tokens
    """

    def __init__(self):
        self.prediction = None
        self.confidence = None
        self.num_op = 0

    def post_processing(self):
        raise NotImplementedError

    def post_process(self, prediction, confidence_score=None):
        """
        Apply dataset-specific post-processing
        """
        self.prediction = list(prediction)
        self.confidence = (
            list(confidence_score) if confidence_score is not None else None
        )
        if self.confidence is not None:
            assert len(self.prediction) == len(self.confidence)
        return self.post_processing()

    def insert_label(self, index, label):
        """
        Insert token at specific index. The associated confidence score is set to 0.
        """
        self.prediction.insert(index, label)
        if self.confidence is not None:
            self.confidence.insert(index, 0)
        self.num_op += 1

    def del_label(self, index):
        """
        Remove the token at a specific index.
        """
        del self.prediction[index]
        if self.confidence is not None:
            del self.confidence[index]
        self.num_op += 1


class PostProcessingModuleSIMARA(PostProcessingModule):
    """
    Specific post-processing for the SIMARA dataset at page level
    """

    def __init__(self):
        super(PostProcessingModuleSIMARA, self).__init__()
        self.matching_tokens = SIMARA_MATCHING_TOKENS
        self.reverse_matching_tokens = dict()
        for key in self.matching_tokens:
            self.reverse_matching_tokens[self.matching_tokens[key]] = key

    def post_processing(self):
        ind = 0
        begin_token = None
        while ind != len(self.prediction):
            char = self.prediction[ind]
            # a tag must be closed before starting a new one
            if char in self.matching_tokens.keys():
                if begin_token is None:
                    ind += 1
                else:
                    self.insert_label(ind, self.matching_tokens[begin_token])
                    ind += 2
                begin_token = char
                continue
            # an end token without prior corresponding begin token is removed
            elif char in self.matching_tokens.values():
                if begin_token == self.reverse_matching_tokens[char]:
                    ind += 1
                    begin_token = None
                else:
                    self.del_label(ind)
                continue
            else:
                ind += 1
        # a tag must be closed
        if begin_token is not None:
            self.insert_label(ind + 1, self.matching_tokens[begin_token])
        res = "".join(self.prediction)
        if self.confidence is not None:
            return res, np.array(self.confidence)
        return res
