# -*- coding: utf-8 -*-

import logging
import pickle

import cv2
import numpy as np
import torch
import yaml

from dan.decoder import GlobalHTADecoder
from dan.models import FCN_Encoder
from dan.ocr.utils import LM_ind_to_str


class DAN:
    """
    The DAN class is used to apply a DAN model.
    The class initializes useful parameters: the device.
    """

    def __init__(self, device):
        """
        Constructor of the DAN class.
        :param device: The device to use.
        """
        super(DAN, self).__init__()
        self.device = device

    def load(self, model_path, params_path, charset_path, mode="eval"):
        """
        Load a trained model.
        :param model_path: Path to the model.
        :param params_path: Path to the parameters.
        :param charset_path: Path to the charset.
        :param mode: The mode to load the model (train or eval).
        """
        with open(params_path, "r") as f:
            parameters = yaml.safe_load(f)["parameters"]
            parameters["decoder"]["device"] = self.device

        with open(charset_path, "rb") as f:
            self.charset = pickle.load(f)

        # Restore the model weights.
        checkpoint = torch.load(model_path, map_location=self.device)

        encoder = FCN_Encoder(parameters["encoder"]).to(self.device)
        encoder.load_state_dict(checkpoint["encoder_state_dict"], strict=True)

        decoder = GlobalHTADecoder(parameters["decoder"]).to(self.device)
        decoder.load_state_dict(checkpoint["decoder_state_dict"], strict=True)
        logging.debug(f"Loaded model {model_path}")

        if mode == "train":
            encoder.train()
            decoder.train()
        elif mode == "eval":
            encoder.eval()
            decoder.eval()
        else:
            raise Exception("Unsupported mode")

        self.encoder = encoder
        self.decoder = decoder
        self.mean, self.std = parameters["mean"], parameters["std"]
        self.max_chars = parameters["max_char_prediction"]

    def predict(self, input_image, confidences=False):
        """
        Run prediction on an input image.
        :param input_image: The image to predict.
        :param confidences: Return the characters probabilities.
        """
        # Preprocess image.
        assert isinstance(
            input_image, np.ndarray
        ), "Input image must be an np.array in RGB"
        input_image = np.asarray(input_image)
        if len(input_image.shape) < 3:
            input_image = cv2.cvtColor(input_image, cv2.COLOR_GRAY2RGB)

        reduced_size = [input_image.shape[:2]]

        input_image = (input_image - self.mean) / self.std
        input_image = np.expand_dims(input_image.transpose((2, 0, 1)), axis=0)
        input_tensor = torch.from_numpy(input_image).to(self.device)
        logging.debug("Image pre-processed")

        start_token = len(self.charset) + 1
        end_token = len(self.charset)

        # Run the prediction.
        with torch.no_grad():
            b = input_tensor.size(0)
            reached_end = torch.zeros((b,), dtype=torch.bool, device=self.device)
            prediction_len = torch.zeros((b,), dtype=torch.int, device=self.device)
            predicted_tokens = (
                torch.ones((b, 1), dtype=torch.long, device=self.device) * start_token
            )
            predicted_tokens_len = torch.ones((b,), dtype=torch.int, device=self.device)

            whole_output = list()
            confidence_scores = list()
            cache = None
            hidden_predict = None

            features = self.encoder(input_tensor.float())
            features_size = features.size()
            coverage_vector = torch.zeros(
                (features.size(0), 1, features.size(2), features.size(3)),
                device=self.device,
            )
            pos_features = self.decoder.features_updater.get_pos_features(features)
            features = torch.flatten(pos_features, start_dim=2, end_dim=3).permute(
                2, 0, 1
            )
            enhanced_features = pos_features
            enhanced_features = torch.flatten(
                enhanced_features, start_dim=2, end_dim=3
            ).permute(2, 0, 1)

            for i in range(0, self.max_chars):
                output, pred, hidden_predict, cache, weights = self.decoder(
                    features,
                    enhanced_features,
                    predicted_tokens,
                    reduced_size,
                    predicted_tokens_len,
                    features_size,
                    start=0,
                    hidden_predict=hidden_predict,
                    cache=cache,
                    num_pred=1,
                )
                whole_output.append(output)
                confidence_scores.append(
                    torch.max(torch.softmax(pred[:, :], dim=1), dim=1).values
                )
                coverage_vector = torch.clamp(coverage_vector + weights, 0, 1)
                predicted_tokens = torch.cat(
                    [
                        predicted_tokens,
                        torch.argmax(pred[:, :, -1], dim=1, keepdim=True),
                    ],
                    dim=1,
                )
                reached_end = torch.logical_or(
                    reached_end, torch.eq(predicted_tokens[:, -1], end_token)
                )
                predicted_tokens_len += 1

                prediction_len[reached_end == False] = i + 1  # noqa E712

                if torch.all(reached_end):
                    break

            confidence_scores = (
                torch.cat(confidence_scores, dim=1).cpu().detach().numpy()
            )
            predicted_tokens = predicted_tokens[:, 1:]
            prediction_len[torch.eq(reached_end, False)] = self.max_chars - 1
            predicted_tokens = [
                predicted_tokens[i, : prediction_len[i]] for i in range(b)
            ]
            confidence_scores = [
                confidence_scores[i, : prediction_len[i]].tolist() for i in range(b)
            ]
            predicted_text = [
                LM_ind_to_str(self.charset, t, oov_symbol="") for t in predicted_tokens
            ]
            logging.info("Image processed")

        if confidences:
            return predicted_text[0], confidence_scores[0]
        return predicted_text[0]
