# -*- coding: utf-8 -*-
import cv2
import numpy as np
import torch
from torch.distributions.uniform import Uniform


def randint(low, high):
    """
    call torch.randint to preserve random among dataloader workers
    """
    return int(torch.randint(low, high, (1,)))


def rand():
    """
    call torch.rand to preserve random among dataloader workers
    """
    return float(torch.rand((1,)))


def rand_uniform(low, high):
    """
    call torch uniform to preserve random among dataloader workers
    """
    return float(Uniform(low, high).sample())


def pad_sequences_1D(data, padding_value):
    """
    Pad data with padding_value to get same length
    """
    x_lengths = [len(x) for x in data]
    longest_x = max(x_lengths)
    padded_data = np.ones((len(data), longest_x)).astype(np.int32) * padding_value
    for i, x_len in enumerate(x_lengths):
        padded_data[i, :x_len] = data[i][:x_len]
    return padded_data


def resize_max(img, max_width=None, max_height=None):
    if max_width is not None and img.shape[1] > max_width:
        ratio = max_width / img.shape[1]
        new_h = int(np.floor(ratio * img.shape[0]))
        new_w = int(np.floor(ratio * img.shape[1]))
        img = cv2.resize(img, (new_w, new_h), interpolation=cv2.INTER_LINEAR)
    if max_height is not None and img.shape[0] > max_height:
        ratio = max_height / img.shape[0]
        new_h = int(np.floor(ratio * img.shape[0]))
        new_w = int(np.floor(ratio * img.shape[1]))
        img = cv2.resize(img, (new_w, new_h), interpolation=cv2.INTER_LINEAR)
    return img


def pad_images(data, padding_value, padding_mode="br"):
    """
    data: list of numpy array
    mode: "br"/"tl"/"random" (bottom-right, top-left, random)
    """
    x_lengths = [x.shape[0] for x in data]
    y_lengths = [x.shape[1] for x in data]
    longest_x = max(x_lengths)
    longest_y = max(y_lengths)
    padded_data = (
        np.ones((len(data), longest_x, longest_y, data[0].shape[2])) * padding_value
    )
    for i, xy_len in enumerate(zip(x_lengths, y_lengths)):
        x_len, y_len = xy_len
        if padding_mode == "br":
            padded_data[i, :x_len, :y_len, ...] = data[i]
        elif padding_mode == "tl":
            padded_data[i, -x_len:, -y_len:, ...] = data[i]
        elif padding_mode == "random":
            xmax = longest_x - x_len
            ymax = longest_y - y_len
            xi = randint(0, xmax) if xmax >= 1 else 0
            yi = randint(0, ymax) if ymax >= 1 else 0
            padded_data[i, xi : xi + x_len, yi : yi + y_len, ...] = data[i]
        else:
            raise NotImplementedError("Undefined padding mode: {}".format(padding_mode))
    return padded_data


def pad_image(
    image,
    padding_value,
    new_height=None,
    new_width=None,
    pad_width=None,
    pad_height=None,
    padding_mode="br",
    return_position=False,
):
    """
    data: list of numpy array
    mode: "br"/"tl"/"random" (bottom-right, top-left, random)
    """
    if pad_width is not None and new_width is not None:
        raise NotImplementedError("pad_with and new_width are not compatible")
    if pad_height is not None and new_height is not None:
        raise NotImplementedError("pad_height and new_height are not compatible")

    h, w, c = image.shape
    pad_width = (
        pad_width
        if pad_width is not None
        else max(0, new_width - w)
        if new_width is not None
        else 0
    )
    pad_height = (
        pad_height
        if pad_height is not None
        else max(0, new_height - h)
        if new_height is not None
        else 0
    )

    if not (pad_width == 0 and pad_height == 0):
        padded_image = np.ones((h + pad_height, w + pad_width, c)) * padding_value
        if padding_mode == "br":
            hi, wi = 0, 0
        elif padding_mode == "tl":
            hi, wi = pad_height, pad_width
        elif padding_mode == "random":
            hi = randint(0, pad_height) if pad_height >= 1 else 0
            wi = randint(0, pad_width) if pad_width >= 1 else 0
        else:
            raise NotImplementedError("Undefined padding mode: {}".format(padding_mode))
        padded_image[hi : hi + h, wi : wi + w, ...] = image
        output = padded_image
    else:
        hi, wi = 0, 0
        output = image

    if return_position:
        return output, [[hi, hi + h], [wi, wi + w]]
    return output


def pad_image_width_right(img, new_width, padding_value):
    """
    Pad img to right side with padding value to reach new_width as width
    """
    h, w, c = img.shape
    pad_width = max((new_width - w), 0)
    pad_right = np.ones((h, pad_width, c), dtype=img.dtype) * padding_value
    img = np.concatenate([img, pad_right], axis=1)
    return img
